/*
Copyright (c) 2019-2025,
Lawrence Livermore National Security, LLC;
See the top-level NOTICE for additional details. All rights reserved.
SPDX-License-Identifier: BSD-3-Clause
*/
#pragma once

#include "gtest/gtest.h"

namespace test {
static const double tolerance = 1e-6;
static const double precise_tolerance = 1e-12;
}  // namespace test
