.wy-side-nav-search {
  display: block;
  width: 300px;
  padding: 0.809em;
  margin-bottom: 0.809em;
  z-index: 200;
  background-color: #fcfcfc;
  text-align: center;
  padding: 0.809em;
  display: block;
  color: #fcfcfc;
  margin-bottom: 0.809em;
}
